__all__ = ["curlometer"]

import numpy as np


def curlometer(Q1, Q2, Q3, Q4, R1, R2, R3, R4):
    r"""
    Calculate the divergence and curl of a vector field from 4-point
    measurements using the curlometer technique. For reference see
    Chapter 14 of Paschmann & Daly (2000), "Analysis Methods for
    Multi-Spacecraft Data", ISSI Scientific Report, SR-001.

    Parameters
    ----------
    Q1, Q2, Q3, Q4: array_like, shape (..., 3)
        Values of vector field at 4 spatial positions. The last axis is
        vector components
    R1, R2, R3, R4: array_like, shape (..., 3)
        Position of the 4 measurements. The last axis is vector
        components

    Return
    ------
    curlometer: dict
        Contains div_Q, curl_Q
    """

    dR_12 = R2 - R1
    dR_13 = R3 - R1
    dR_14 = R4 - R1

    k2 = (tmp := np.cross(dR_13, dR_14, axis=-1)) / np.einsum(
        "...i,...i", dR_12, tmp
    )[:, np.newaxis]
    k3 = (tmp := np.cross(dR_12, dR_14, axis=-1)) / np.einsum(
        "...i,...i", dR_13, tmp
    )[:, np.newaxis]
    k4 = (tmp := np.cross(dR_12, dR_13, axis=-1)) / np.einsum(
        "...i,...i", dR_14, tmp
    )[:, np.newaxis]
    k1 = -k2 - k3 - k4

    div_Q = np.sum(k1 * Q1 + k2 * Q2 + k3 * Q3 + k4 * Q4, axis=-1)
    curl_Q = (
        np.cross(k1, Q1, axis=-1)
        + np.cross(k2, Q2, axis=-1)
        + np.cross(k3, Q3, axis=-1)
        + np.cross(k4, Q4, axis=-1)
    )

    return dict(div_Q=div_Q, curl_Q=curl_Q)



__all__ = ["_setup_matplotlib"]

import atexit
import os
import shutil
import logging
import tempfile
from importlib.resources import files

import matplotlib as mpl
import matplotlib.pyplot as plt

_latex_binary = shutil.which("latex")
_stylesheet = files("tvolib.matplotlib_utils") / "data" / "tvolib.mplstyle"


def _setup_matplotlib(use_tex=False, cache_tex=False):
    r"""
    Configure matplotlib with custom stylesheet and settings

    Parameter
    ---------
    use_tex: bool
        Enable latex if toggled and a binary is found in the system path
    cache_tex: bool
        Configure caching behavior of matplotlib for latex outputs.
        Recommended for multiprocessing
    """

    logging.info("Using tvolib stylesheet")
    plt.style.use(_stylesheet)
    plt.rc("text", usetex=use_tex and _latex_binary is not None)

    if cache_tex:
        # Quick fix for matplotlib's tex cache in multiprocessing
        mpldir = tempfile.mkdtemp()
        atexit.register(shutil.rmtree, mpldir)
        umask = os.umask(0)
        os.umask(umask)
        os.chmod(mpldir, 0o777 & ~umask)
        os.environ["HOME"] = mpldir
        os.environ["MPLCONFIGDIR"] = mpldir

        class TexManager(mpl.texmanager.TexManager):
            texcache = os.path.join(mpldir, "tex.cache")

        mpl.texmanager.TexManager = TexManager



__all__ = [
    "add_colorbar",
    "add_panel_label",
    "format_datetime_labels",
]

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


def add_colorbar(
    ax: plt.Axes,
    where: str = "right",
    pad: float = 0.05,
    size: str = "2%",
) -> plt.Axes:
    r"""
    Adds colorbar next to a given `Axes` instance. See `matplotlib`
    documentation for `where`, `pad`, and `size` options.

    Parameters
    ----------
    ax: `matplotlib.axes.Axes`
        The `Axes` instance to draw on.
    where: str
        Where to append the colorbar with respective to the `Axes`.
    pad: str
        Padding of the colorbar from the `Axes`.
    size: str
        Size of the colorbar relative to the `Axes`.

    Return
    ------
    cax: `matplotlib.axes.Axes`
        A child `Axes` instance created relative to the input `Axes` for
        drawing the colorbar.
    """
    divider = make_axes_locatable(ax)
    return divider.append_axes(where, size=size, pad=pad)


def add_panel_label(
    ax: plt.Axes,
    x: float,
    y: float,
    text: str,
    box_format: dict = dict(facecolor="wheat", alpha=0.9),
    **kwargs,
) -> None:
    r"""
    Adds a boxed text to an `Axes` instance. Intended for drawing panel
    annotations for a figure.

    Parameters
    ----------
    ax: `matplotlib.axes.Axes`
        The `Axes` instance to draw on.
    x, y: float
        Location of the text.
    text: str
        Content of the the text.
    box_format: dict
        Formatting of the enclosing box.
    kwargs: dict
        Other options for `Axes.text` method.
    """
    ax.text(x, y, text, transform=ax.transAxes, bbox=box_format, **kwargs)


def format_datetime_labels(ax: plt.Axes) -> None:
    r"""
    Format the labels for plots with `numpy.datetime64` abscissa.

    Parameter
    ---------
    ax: `matplotlib.axes.Axes`
        The `Axes` instance.
    """
    locator = mdates.AutoDateLocator()
    formatter = mdates.ConciseDateFormatter(locator)
    ax.xaxis.set_major_formatter(formatter)



__all__ = ["sampling_period", "sampling_frequency"]

import astropy.units as u
import numpy as np


def sampling_period(t):
    r"""Calculate the sampling period of a time array."""

    if not np.issubdtype(t.dtype, np.datetime64):
        raise NotImplementedError("Input must be datetime64 dtype.")

    t = t.astype("datetime64[ns]")
    return (
        np.diff(t).mean().astype("timedelta64[ns]").astype("f8") * u.ns
    ).to(u.s)


def sampling_frequency(t):
    r"""Calculate the sampling frequency of a time array."""
    return (1 / sampling_period(t)).to(u.Hz)



__all__ = ["Lin10MagnetopauseModel"]

import numpy as np
from scipy.interpolate import UnivariateSpline


class Lin10MagnetopauseModel:
    r"""
    3D asymmetric magnetopause model. For references consult
    Lin+ (2010), "A three-dimensional asymmetric magnetopause model",
    JGR-Space Physics, doi:10.1029/2009JA014235
    """

    # Fit coefficients (see Table 9 in reference)
    a = np.zeros(22, dtype=np.float32)
    a[0] = 12.544
    a[1] = -0.194
    a[2] = 0.305
    a[3] = 0.0573
    a[4] = 2.178
    a[5] = 0.0571
    a[6] = -0.999
    a[7] = 16.473
    a[8] = 0.00152
    a[9] = 0.382
    a[10] = 0.0431
    a[11] = -0.00763
    a[12] = -0.210
    a[13] = 0.0405
    a[14] = -4.430
    a[15] = -0.636
    a[16] = -2.600
    a[17] = 0.832
    a[18] = -5.328
    a[19] = 1.103
    a[20] = -0.907
    a[21] = 1.450

    def __init__(self, pressure, bfield, tilt_angle):
        r"""

        Parameters
        ----------
        pressure: nPa
            Total (dynamic and magnetic) solar wind pressure
        bfield: nT
            z coordinate of the interplanetary magnetic field
        tilt_angle: rad
            Dipole tilt angle
        """

        # Save
        self._P = pressure
        self._Bz = bfield
        self._phi = tilt_angle

        # Calculate other coefficients
        self.construct_model()

    def construct_model(self):
        r"""
        Calculate the coefficients in eq. (20) and r(theta, varphi)
        in eq. (19)
        """
        # Unpack
        a = self.a
        P = self._P
        Bz = self._Bz
        phi = self._phi

        # Auxilliary coefficients
        r0 = (
            a[0]
            * (P ** a[1])
            * (1 + a[2] * (np.exp(a[3] * Bz) - 1) / (np.exp(a[4] * Bz) + 1))
        )
        b0 = a[6] + a[7] * (np.exp(a[8] * Bz) - 1) / (np.exp(a[9] * Bz) + 1)
        b1 = a[10]
        b2 = a[11] + a[12] * phi
        b3 = a[13]
        cn = cs = a[14] * (P ** a[15])
        dn = a[16] + a[17] * phi + a[18] * (phi**2)
        ds = a[16] - a[17] * phi + a[18] * (phi**2)
        en = es = a[21]
        tn = a[19] + a[20] * phi
        ts = a[19] - a[20] * phi

        # Auxilliary functions
        def psi_n(varphi, theta):
            return np.arccos(
                np.cos(theta) * np.cos(tn)
                + np.sin(theta) * np.sin(tn) * np.cos(varphi - np.pi / 2)
            )

        def psi_s(varphi, theta):
            return np.arccos(
                np.cos(theta) * np.cos(ts)
                + np.sin(theta) * np.sin(ts) * np.cos(varphi - 3 * np.pi / 2)
            )

        def f(varphi, theta):
            return (
                np.cos(theta / 2)
                + a[5] * np.sin(theta * 2) * (1 - np.exp(-theta))
            ) ** (
                b0
                + b1 * np.cos(varphi)
                + b2 * np.sin(varphi)
                + b3 * (np.sin(varphi) ** 2)
            )

        # Radial function
        def r(varphi, theta):
            return (
                r0 * f(varphi, theta)
                + cn * np.exp(dn * (psi_n(varphi, theta) ** en))
                + cs * np.exp(ds * (psi_s(varphi, theta) ** es))
            )

        self.radial_distance = r

    def shape_N(self, which="XY", N=1000, rotation_angle=0):
        theta = np.linspace(0, np.pi, N // 2)
        if which == "XY":
            # Get negative side of Y_gsm
            Rm = self.radial_distance(np.pi, theta[::-1])
            Xm = Rm * np.cos(theta[::-1])
            Ym = -Rm * np.sin(theta[::-1])
            # Get positive side of Y_gsm
            Rp = self.radial_distance(0, theta[1:])
            Xp = Rp * np.cos(theta[1:])
            Yp = Rp * np.sin(theta[1:])
            # Combine
            X = np.append(Xm, Xp)
            Y = np.append(Ym, Yp)
            # Rotate
            R = np.sqrt(X**2 + Y**2)
            T = np.arctan2(Y, X) + rotation_angle
            X = R * np.cos(T)
            Y = R * np.sin(T)
            return X, Y
        elif which == "XZ":
            # Get negative side of Z_gsm
            Rm = self.radial_distance(-np.pi / 2, theta[::-1])
            Xm = Rm * np.cos(theta[::-1])
            Zm = -Rm * np.sin(theta[::-1])
            # Get positive side of Z_gsm
            Rp = self.radial_distance(np.pi / 2, theta[1:])
            Xp = Rp * np.cos(theta[1:])
            Zp = Rp * np.sin(theta[1:])
            # Combine
            X = np.append(Xm, Xp)
            Z = np.append(Zm, Zp)
            # Rotate
            R = np.sqrt(X**2 + Z**2)
            T = np.arctan2(-Z, X) + rotation_angle
            X = R * np.cos(T)
            Z = R * np.sin(T)
            return X, Z
        else:
            raise NotImplementedError

    def shape_F(self, which="XY", N=1000000, rotation_angle=0):
        if which == "XY":
            X, Y = self.shape_N(which, N, rotation_angle=rotation_angle)
            R = np.sqrt(X**2 + Y**2)
            T = np.arctan2(Y, X)
        elif which == "XZ":
            X, Z = self.shape_N(which, N, rotation_angle=rotation_angle)
            R = np.sqrt(X**2 + Z**2)
            T = np.arctan2(-Z, X)
        else:
            raise NotImplementedError

        idx = np.argsort(T)
        return UnivariateSpline(T[idx], R[idx], s=0)

    def add_mpause_model(self, ax, which="XY", rotation_angle=0, **kw):
        ax.plot(
            *self.shape_N(which=which, rotation_angle=rotation_angle), **kw
        )

    @property
    def pressure(self):
        return self._P

    @property
    def bfield(self):
        return self._Bz

    @property
    def tilt_angle(self):
        return self._phi

    @pressure.setter
    def pressure(self, new_pressure):
        self._P = new_pressure

    @bfield.setter
    def bfield(self, new_bfield):
        self._Bz = new_bfield

    @tilt_angle.setter
    def tilt_angle(self, new_tilt_angle):
        self._phi = new_tilt_angle


if __name__ == "__main__":
    from tvolib import mpl_utils as mu

    model = Lin10MagnetopauseModel(pressure=1, bfield=0, tilt_angle=0)
    X_gsm, Y_gsm = model.shape_N(which="XY")
    T = np.linspace(-np.pi, np.pi, 1000)
    f = model.shape_F(which="XY", rotation_angle=np.radians(0))
    R = f(T)

    fig, ax = mu.plt.subplots(1, 1)

    mu.draw_earth(ax)
    ax.plot(X_gsm, Y_gsm, "-k")
    ax.plot(R * np.cos(T), R * np.sin(T), "--r")
    ax.set_xlim(-40, 20)
    ax.set_ylim(-35, 35)
    ax.set_aspect("equal")

    mu.plt.show()



import colorcet
import matplotlib as mpl

from ._draw import draw_arrows, draw_earth, draw_multicolored_line
from ._format import add_colorbar, add_panel_label, format_datetime_labels
from ._setup import _setup_matplotlib

_setup_matplotlib()

# Aliases
mplc = mpl.colors
plt = mpl.pyplot



__all__ = ["_setup_logger"]

import logging
import sys
from os import environ


def _setup_logger():
    r"""Configure logging format for stdout"""

    logging.basicConfig(
        format=r"%(asctime)s [%(levelname)s]: %(message)s",
        level=logging.INFO if environ.get("DEBUG") is None else logging.DEBUG,
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    logging.captureWarnings(True)



__all__ = [
    "draw_arrows",
    "draw_earth",
    "draw_multicolored_line",
]

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt


def draw_arrows(
    ax: plt.Axes,
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    number_of_arrows: int | None = None,
    color: str = "k",
) -> None:
    r"""
    Draws arrows along a path on a given `Axes` instance

    Parameters
    ----------
    ax: `matplotlib.axes.Axes`
        The `Axes` instance to draw on.
    x: array_like, shape (n,)
        1D array containing the path's x-coordinates.
    y: array_like, shape (n,)
        1D array containing the path's y-coordinates.
    number_of_arrows: int
        Number of arrows to draw.
    color: str
        Arrow color
    """

    # Pre-process inputs and sanity checks
    x = np.array(x)
    y = np.array(y)
    assert x.ndim == y.ndim == 1, "Input `x` and `y` coordinates must be 1D"

    # Calculate arrow spacing along the path and where to draw them
    data_size = len(x)
    number_of_arrows = (
        data_size // 5 if number_of_arrows is None else number_of_arrows
    )
    spacing = data_size // (number_of_arrows + 1)
    draw_idx = np.arange(spacing, data_size, spacing)

    # Draw arrows on patches
    for i in draw_idx:
        arrow_patch = mpl.patches.FancyArrowPatch(
            (x[i - 1], y[i - 1]),
            (x[i], y[i]),
            arrowstyle="->",
            mutation_scale=20,
            color=color,
        )
        ax.add_patch(arrow_patch)


def draw_earth(
    ax: plt.Axes,
    radius: float = 1,
    number_of_points: int = 50,
    zorder: int = 999,
) -> None:
    r"""
    Draws the Earth on a given `Axes` instance with shading indicating
    day/night. Dayside is x>0.

    Parameters
    ----------
    ax: `matplotlib.axes.Axes`
        The `Axes` instance to draw on.
    radius: float
        The radius of the Earth on the `Axes`.
    number_of_points: int
        The larger the number of points, the smoother the drawn shape.
    zorder: int
        Order on the `Axes`.
    """

    # Nightside
    theta = np.linspace(np.pi / 2, 3 * np.pi / 2, number_of_points)
    Xn = radius * np.cos(theta)
    Yn = radius * np.sin(theta)
    Xn = np.append(Xn, Xn[0])
    Yn = np.append(Yn, Yn[0])

    # Dayside
    theta = np.linspace(-np.pi / 2, np.pi / 2, number_of_points)
    Xd = radius * np.cos(theta)
    Yd = radius * np.sin(theta)
    Xd = np.append(Xd, Xd[0])
    Yd = np.append(Yd, Yd[0])

    # Plot
    ax.plot(Xd, Yd, "-k", zorder=zorder)
    ax.plot(Xn, Yn, "-k", zorder=zorder)
    ax.fill(Xd, Yd, color="w")
    ax.fill(Xn, Yn, color="k")


def draw_multicolored_line(
    ax: plt.Axes,
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    c: npt.ArrayLike,
    cmap: str = "cet_rainbow",
    vmin: float | None = None,
    vmax: float | None = None,
    **kwargs,
) -> mpl.collections.Collection:
    r"""
    Draw a colored line on a given `Axes` instance

    Parameters
    ----------
    ax: `matplotlib.axes.Axes`
        The `Axes` instance to draw on.
    x: array_like, shape (n,)
        1D array containing the x-coordinates.
    y: array_like, shape (n,)
        1D array containing the y-coordinates.
    c: array_like, shape (n,)
        1D array containing the colored data.
    cmap: str

    Return
    ------
    line: `mpl.collections.LineCollection`
        The line collection instance created from drawing all the colored
        segments.
    """

    # Pre-process inputs and sanity checks
    x = np.array(x)
    y = np.array(y)
    c = np.array(c)
    vmin = c.min() if vmin is None else vmin
    vmax = c.max() if vmax is None else vmax
    assert (
        x.ndim == y.ndim == c.ndim == 1
    ), "Input `x` and `y` coordinates and colored data `c` must be 1D"

    # Create points and segments from coordinates
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    # Create line segments and color them based on `c`
    lc = mpl.collections.LineCollection(
        segments,
        cmap=cmap,
        norm=mpl.colors.Normalize(vmin, vmax),
        **kwargs,
    )
    lc.set_array(c)
    line = ax.add_collection(lc)

    # Automatically set limits to axis
    if len(ax.get_lines()) == 1:
        Lx = x.max() - x.min()
        Ly = y.max() - y.min()
        ax.set_xlim(x.min() - 0.1 * Lx, x.max() + 0.1 * Lx)
        ax.set_ylim(y.min() - 0.1 * Ly, y.max() + 0.1 * Ly)

    return line
